### MM Algorithm for Rank Regression
###
### Brent Johnson
### 9 January 2008
###
### y = n-by-1 response vector
### x = n-by-p matrix of predictors
### binit = p-by-1 vector of initial values (full R-estimates)
### lambda = regularization parameter
### awt = lasso weights
### eps = an effective zero
###
### NOTES : we use two functions supplied by other authors
### namely, Jeff Terpstra (NDSU, Stat) & Joe McKean (WMU, Stat)
### wilcoxontau() : calculates tau for updating step of hessian
### pairup() : internal function
###
rankmm <- function(y,x,binit,lambda,awt=rep(1,length(binit)),
	eps=.Machine$double.eps,MAXITER=100,TOL=1E-8) {
	
	xtx <- t(x) %*% x
	p <- length(binit)
	n <- length(y)
	it <- 1
	absdiff <- 10
	bnew <- binit
	while ((it<MAXITER) && (absdiff>TOL)) {
		bold <- as.vector(bnew)
		cat("b = ",round(bold,4),"\n")
		res <- y - x %*% bold
		wt <- (rank(res)/(n+1)) - 0.5
		Sorig <- -1*(t(x) %*% wt)
		nSig <- n*lambda*diag(awt/(1E-4 + abs(bold)))
		pen <- n*lambda*awt*sign(bold) 	# derivative of penalty function
		#pen <- nSig %*% bold			# quadratic approximation to derivative
		S <- Sorig + pen 
		tau <- wilcoxontau(res,p)
		A <- xtx/tau
		H <- A+nSig
		bstep <- solve(H,S)
		bnew <- bold - bstep
		it <- it + 1
		absdiff <- max(abs(bold-bnew)) 
		
	}
	
	bhat <- as.vector(bnew)
	zero.vec <- bhat < eps # declare effective zero
	bhat[zero.vec] <- 0
	HInv <- solve(H)
	V <- HInv %*% xtx %*% HInv
	sehat <- sqrt(diag(V))
	list(b=bhat,se=sehat)
}

# R Functions for Weighted Wilcoxon Estimation and Inference  
#
# Version:      1.3
# Last Update:  05/30/06
# Requirements: 'quantreg' package
#               'lqs' package (R 1.8.1 or lower)
#               'MASS' package (R 1.9.0 or higher)
# 
# Authors:      Jeff Terpstra                      Joe McKean
# 
# Address:      Department of Statistics           Department of Statistics
#               North Dakota State University      Western Michigan University
#               P.O. Box 5575 Waldron 201H         5506 Everett Tower
#               Fargo, ND  58105                   Kalamazoo, MI  49008
#
# Email:        Jeff.Terpstra@ndsu.nodak.edu       joe@stat.wmich.edu
#
# Phone:        701-231-8188                       269-387-4541
#
# Fax:          701-231-8734                       269-387-1419                     
#
# Notes:        These functions require the 'quantreg' and 'MASS'/'lqs' packages
#               which are available from http://cran.r-project.org/
#               For example,
#                 -install.packages("quantreg",lib="/destdir") 
#                 -library("quantreg",lib.loc="./") loads the "quantreg" package
#                  on a UNIX machine
#                 -library("MASS") loads the "MASS" package
#

#
# Function that sets up pairwise comparisons
#

pairup=function(x,type="less") {
  x=as.matrix(x)
  n=dim(x)[1]
  i=rep(1:n,rep(n,n))
  j=rep(1:n,n)
  c1=apply(x,2,function(y){rep(y,rep(length(y),length(y)))})
  c2=apply(x,2,function(y){rep(y,length(y))})
  ans=cbind(c1,c2)
  ans=switch(type, less=ans[(i<j), ], leq=ans[i<=j, ], neq=ans)
  ans
}


###############################################################################
#                                                                             #
# Functions that calculate scale parameter estimates                          #
#                                                                             #
###############################################################################
#
# Density estimate of wilcoxontau = 1/(sqrt(12)*\int f^2(x)dx).
# Discussed in Section 3.7.1 of HM.
#
# note:  Use delta=0.80 if n/p > 5 and delta=0.95 otherwise (HM (p.184)).
# note:  This is Joe's version (from joe3.r) which contains a degrees of
#        freedom correction, Huber's correction (Huber, p.174), and a
#        standardization.
#
#  resd: is the vector of residuals.
#     p: is the number of regression coefficients (without the intercept), but
#        we use p+1 for the DF/Huber corrections.
# delta: Bandwidth parameter for estimate of tau.   Values  should be 
#        between .8 and .99.  Larger values result in a more conservative test.
#        If n/p is less than 5, recommend delta = .95.   Default is .8.
# param: This is a parameter for Huber's degrees of freedom correcton
#        as discussed in Huber (1981, p.174) and it is used in the estimator
#        of tau.  Smaller values lead to more conservative estimates of tau.
#        Default is 2.   It is the benchmark for labeling a standardized
#        residual as a potential outlier.
#
wilcoxontau<-function(resd,p,delta=if((length(resd)/p) > 5) 0.80 else 0.95,param=2){
  eps<-.000001
  n<-length(resd)
  temp<-pairup(resd)
  dresd<-sort(abs(temp[,1]-temp[,2]))
  dresd = dresd[(p+1):choose(n,2)]
  tdeltan<-quantile(dresd,delta)/sqrt(n)
  w<-rep(0,length(dresd))
  w[dresd <= tdeltan]<-1
  cn<-2/(n*(n-1))
  scores = sqrt(12)*((1:n)/(n+1) - .5)
  mn = mean(scores)
  con = sqrt(sum((scores - mn)^2)/(n+1))
  scores = (scores - mn)/con
  dn = scores[n] - scores[1]
  wilcoxontau<-sqrt(n/(n-p-1))*((2*tdeltan)/(dn*sum(w)*cn))
  w<-rep(0,n)
  stan<-(resd-median(resd))/mad(resd)
  w[abs(stan) < param]<-1
  hubcor<-sum(w)/n
  if(hubcor < eps){hubcor<-eps}
  fincor<-1 + (((p+1)/n)*((1-hubcor)/hubcor))
  wilcoxontau<-fincor*wilcoxontau
  names(wilcoxontau)<-NULL
  wilcoxontau
}
